namespace Murphy.AutoDiff
open System

/// Floating point expression with automatic differentiation support.
[<AbstractClass>]
type DFloat<[<Measure>] 'U> private (value:float<'U>) =
  /// Change the unit of measure, for cases where compiler type inference breaks down :-(
  static let measurecast (x:DFloat<'u>) : DFloat<'v> = x |> box |> unbox

  /// The value of the expression.
  member _.Value = value

  /// The sign of the expression.
  member _.Sign = sign value

  /// Extract the derivative of the expression with respect to the given variable.
  abstract Derive : DFloat<'v> -> DFloat<'U / 'v>

  override _.ToString() = string value

  /// Generic zero constant.
  static member Zero : DFloat<'U> =
    { new DFloat<'U>(0.0<_>) with
        override this.Derive(_) = DFloat.Zero }

  /// Constant expression with zero derivatives.
  static member Val(value:float<'U>) =
    { new DFloat<'U>(value) with
        override _.Derive(_) = DFloat.Zero }

  /// Generic unity constant.
  static member One : DFloat<'U> =
    DFloat.Val(1.0 |> box |> unbox)

  /// Variable expression with unit derivative (if differentiated with respect to itself).
  static member Var(value:float<'U>) =
    { new DFloat<'U>(value) with
        override this.Derive(var) =
          if Object.ReferenceEquals(var, this) then
            DFloat.One
          else
            DFloat.Zero }

  /// Addition implementing the sum rule for derivatives.
  static member (+) (a:DFloat<'U>, b:float<'U>) : DFloat<'U> =
    { new DFloat<'U>(a.Value + b) with
        override _.Derive(var) = a.Derive(var) }

  /// Addition implementing the sum rule for derivatives.
  static member (+) (a:float<'U>, b:DFloat<'U>) : DFloat<'U> =
    { new DFloat<'U>(a + b.Value) with
        override _.Derive(var) = b.Derive(var) }

  /// Addition implementing the sum rule for derivatives.
  static member (+) (a:DFloat<'U>, b:DFloat<'U>) : DFloat<'U> =
    { new DFloat<'U>(a.Value + b.Value) with
        override _.Derive(var) = a.Derive(var) + b.Derive(var) }

  /// Additive inverse implementing the sum rule for derivatives.
  static member (~-) (a:DFloat<'U>) : DFloat<'U> =
    { new DFloat<'U>(-a.Value) with
        override _.Derive(var) = -a.Derive(var) }

  /// Subtraction implementing the sum rule for derivatives.
  static member (-) (a:DFloat<'U>, b:float<'U>) : DFloat<'U> =
    { new DFloat<'U>(a.Value - b) with
        override _.Derive(var) = a.Derive(var) }

  /// Subtraction implementing the sum rule for derivatives.
  static member (-) (a:float<'U>, b:DFloat<'U>) : DFloat<'U> =
    { new DFloat<'U>(a - b.Value) with
        override _.Derive(var) = -b.Derive(var) }

  /// Subtraction implementing the sum rule for derivatives.
  static member (-) (a:DFloat<'U>, b:DFloat<'U>) : DFloat<'U> =
    { new DFloat<'U>(a.Value - b.Value) with
        override _.Derive(var) = a.Derive(var) - b.Derive(var) }

  /// Multiplication implementing the product rule for derivatives.
  static member ( * ) (a:DFloat<'U>, b:float<'v>) : DFloat<'U * 'v> =
    { new DFloat<'U * 'v>(a.Value * b) with
        override _.Derive(var) = a.Derive(var) * b }

  /// Multiplication implementing the product rule for derivatives.
  static member ( * ) (a:float<'v>, b:DFloat<'U>) : DFloat<'v * 'U> =
    { new DFloat<'U * 'v>(a * b.Value) with
        override _.Derive(var) = a * b.Derive(var) }

  /// Multiplication implementing the product rule for derivatives.
  static member ( * ) (a:DFloat<'U>, b:DFloat<'v>) : DFloat<'U * 'v> =
    { new DFloat<'U * 'v>(a.Value * b.Value) with
        override _.Derive(var) = (a.Derive(var) * b) + (a * b.Derive(var |> measurecast) |> measurecast) }

  /// Division implementing the quotient rule for derivatives.
  static member (/) (a:DFloat<'U>, b:float<'v>) : DFloat<'U / 'v> =
    { new DFloat<'U / 'v>(a.Value / b) with
        override _.Derive(var) = a.Derive(var) / b }

  /// Division implementing the quotient rule for derivatives.
  static member (/) (a:float<'v>, b:DFloat<'U>) : DFloat<'v / 'U> =
    { new DFloat<'v / 'U>(a / b.Value) with
        override _.Derive(var) = -a * (b.Derive(var) |> measurecast) / (b * b |> measurecast) }

  /// Division implementing the quotient rule for derivatives.
  static member (/) (a:DFloat<'U>, b:DFloat<'v>) : DFloat<'U / 'v> =
    { new DFloat<'U / 'v>(a.Value / b.Value) with
        override _.Derive(var) = (a.Derive(var) / b) - (a * (b.Derive(var) |> measurecast) / (b * b |> measurecast)) }

  /// Absolute value implementing the chain rule for derivatives.
  static member Abs(x:DFloat<'U>) : DFloat<'U> =
    { new DFloat<'U>(abs x.Value) with
        override _.Derive(var) = x.Derive(var) * double(sign x) }

  /// Exponentiation implementing the power and chain rules for derivatives.
  static member Pow(x:DFloat<1>, p:float) : DFloat<1> =
    { new DFloat<1>(x.Value ** double p) with
        override _.Derive(var) = x.Derive(var) * p * x ** (p - 1.0) }

  /// Exponential function implementing the chain rule for derivatives.
  static member Exp(x:DFloat<1>) : DFloat<1> =
    { new DFloat<1>(exp x.Value) with
        override _.Derive(var) = x.Derive(var) * exp x }

  /// Logarithm function implementing the chain rule for derivatives.
  static member Log(x:DFloat<1>) : DFloat<1> =
    { new DFloat<1>(log x.Value) with
        override _.Derive(var) = x.Derive(var) / x }

  /// Sine function implementing the chain rule for derivatives.
  static member Sin(x:DFloat<1>) : DFloat<1> =
    { new DFloat<1>(sin x.Value) with
        override _.Derive(var) = x.Derive(var) * cos x }

  /// Cosine function implementing the chain rule for derivatives.
  static member Cos(x:DFloat<1>) : DFloat<1> =
    { new DFloat<1>(cos x.Value) with
        override _.Derive(var) = -x.Derive(var) * sin x }

(* vim:set sts=2 ts=4 sw=2 et ai: *)
